@@ -27,6 +27,7 @@
 
 #include "xorg/gtest/xorg-gtest-process.h"
 
+#include <sys/prctl.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <unistd.h>
@@ -42,24 +43,48 @@
 
 struct xorg::testing::Process::Private {
   pid_t pid;
+  enum State state;
 };
 
 xorg::testing::Process::Process() : d_(new Private) {
   d_->pid = -1;
+  d_->state = NONE;
+}
+
+enum xorg::testing::Process::State xorg::testing::Process::GetState() {
+  if (d_->state == RUNNING) {
+    int status;
+    int pid = waitpid(Pid(), &status, WNOHANG);
+    if (pid == Pid()) {
+      if (WIFEXITED(status)) {
+        d_->pid = -1;
+        d_->state = WEXITSTATUS(status) ? FINISHED_FAILURE : FINISHED_SUCCESS;
+      }
+    }
+  }
+
+  return d_->state;
 }
 
 void xorg::testing::Process::Start(const std::string &program, const std::vector<std::string> &argv) {
   if (d_->pid != -1)
     throw std::runtime_error("Attempting to start an already started process");
 
-  d_->pid = vfork();
+  d_->pid = fork();
 
   if (d_->pid == -1) {
+    d_->state = ERROR;
     throw std::runtime_error("Failed to fork child process");
   } else if (d_->pid == 0) { /* Child */
     close(0);
-    close(1);
-    close(2);
+    if (getenv("XORG_GTEST_CHILD_STDOUT") == NULL) {
+      close(1);
+      close(2);
+    }
+
+#ifdef __linux
+    prctl(PR_SET_PDEATHSIG, SIGTERM);
+#endif
 
     std::vector<char*> args;
     std::vector<std::string>::const_iterator it;
@@ -67,23 +92,29 @@ void xorg::testing::Process::Start(const std::string &program, const std::vector
     args.push_back(strdup(program.c_str()));
 
     for (it = argv.begin(); it != argv.end(); it++)
-      if (!it->empty())
-        args.push_back(strdup(it->c_str()));
+      args.push_back(strdup(it->c_str()));
     args.push_back(NULL);
 
     execvp(program.c_str(), &args[0]);
 
+    d_->state = ERROR;
     throw std::runtime_error("Failed to start process");
   }
+
+  d_->state = RUNNING;
 }
 
 void xorg::testing::Process::Start(const std::string& program, va_list args) {
   std::vector<std::string> argv;
 
-  do {
-    std::string arg(va_arg(args, char*));
-    argv.push_back(arg);
-  } while (!argv.back().empty());
+  if (args) {
+    char *arg;
+    do {
+      arg = va_arg(args, char*);
+      if (arg)
+        argv.push_back(std::string(arg));
+    } while (arg);
+  }
 
   Start(program, argv);
 }
@@ -96,14 +127,16 @@ void xorg::testing::Process::Start(const std::string& program, ...) {
 }
 
 bool xorg::testing::Process::WaitForExit(unsigned int timeout) {
-  for (int i = 0; i < 10; i++) {
+  for (unsigned int i = 0; i < timeout * 100; i++) {
     int status;
     int pid = waitpid(Pid(), &status, WNOHANG);
 
     if (pid == Pid())
       return true;
+    else if (pid == -1)
+      return errno == ECHILD;
 
-      usleep(timeout * 100);
+      usleep(10);
   }
 
   return false;
@@ -112,6 +145,19 @@ bool xorg::testing::Process::WaitForExit(unsigned int timeout) {
 bool xorg::testing::Process::KillSelf(int signal, unsigned int timeout) {
   bool wait_success = true;
 
+  enum State state = GetState();
+  switch (state) {
+    case FINISHED_SUCCESS:
+    case FINISHED_FAILURE:
+    case TERMINATED:
+      return true;
+    case ERROR:
+    case NONE:
+      return false;
+    default:
+      break;
+  }
+
   if (d_->pid == -1) {
     return false;
   } else if (d_->pid == 0) {
@@ -120,12 +166,14 @@ bool xorg::testing::Process::KillSelf(int signal, unsigned int timeout) {
   } else { /* Parent */
     if (kill(d_->pid, signal) < 0) {
       d_->pid = -1;
+      d_->state = ERROR;
       return false;
     }
     if (timeout > 0)
       wait_success = WaitForExit(timeout);
     d_->pid = -1;
   }
+  d_->state = TERMINATED;
   return wait_success;
 }
 
